#!/bin/python3

import re
import os
import subprocess
import sys

class build_cmd:
    def __init__(self, gcc, o_file, c_file):
        self.gcc = gcc
        self.o_file = o_file
        self.c_file = c_file
    def print(self):
        print(f"gcc={self.gcc}")    
        print(f"o_file={self.o_file}")    
        print(f"c_file={self.c_file}")    

def print_output(outs):
    print(outs)
        
def exec_cmd(cmd, verbose=False):
    cd_p = re.compile(r"""
        ^\s*cd\s+
        (?P<dir>[^\s]+)\s*$
    """, re.VERBOSE)
    m=cd_p.search(cmd)
    if m:
        dirname=m.group('dir')
        try:
            os.chdir(dirname)
            cur_dir=dirname
            print("OK")
            return ""
        except FileNotFoundError:
            print("error\n%s not found" % (dirname))
            quit(-1)
    complete=subprocess.run(cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, encoding="utf8")
    o=complete.stdout.splitlines()
    if complete.returncode:
        print("error")
        print_output(o)
        quit(-1)
    if verbose:
        print_output(o)
    return o

def collect_build_cmds(dir):
    """
        description:
        ------------
            1. make -C dir clean
            2. make V=1 -C dir                 ### running normal build with verbose on, and collecting gcc_cmds during the process
        outputs:
        --------
            build_cmds list
    """
    build_cmds=[]
    script_dir=os.path.abspath(os.path.dirname(sys.argv[0]))
    o = exec_cmd(f"make -C {dir} PWD={dir} clean")
    o = exec_cmd(f"make -C {dir} PWD={dir} V=1")
    for l in o:
        if not re.compile("^\s*gcc").search(l):
            continue
        m = re.compile("^\s+(?P<gcc_cmd>.*)\s+-c\s+-o\s+(?P<o_file>\S+)\s+(?P<c_file>\S+)$").search(l)
        if m is None:
            continue
        gcc_cmd = m.group('gcc_cmd')
        o_file = m.group('o_file')
        c_file = m.group('c_file')
        if '.mod.c' in c_file:
            continue
        if dir != script_dir:
            local_c_file = c_file.replace(dir,script_dir)
        local_o_file = f"{os.path.splitext(c_file)[0]}.o"
        build_cmds.append(build_cmd(gcc_cmd, local_o_file, c_file))
    return build_cmds

def modify_gcc_from_kernel_to_user(build_cmds_kernel):
    build_cmds_user=[]
    for k in build_cmds_kernel:
        gcc_user = k.gcc
        gcc_user = gcc_user.replace("-fno-PIE", "-fPIC")
        gcc_user = gcc_user.replace("-mcmodel=kernel", "")
        build_cmds_user.append(build_cmd(gcc_user, k.o_file, k.c_file))
    return build_cmds_user

def create_objs_str(build_cmds):
    """
        create objs list
    """
    list_str="\\\n"
    for cmd in build_cmds:
        list_str += f"  {cmd.o_file} \\\n"
    return list_str

def create_compile_rules(build_cmds):
    """
        create compile rule for object files
    """
    rules=""
    for cmd in build_cmds:
        rules+=\
f"""\
{cmd.o_file}:{cmd.c_file}
\tcd /lib/modules/`uname -r`/build; \\
\t{cmd.gcc} -c -o {cmd.o_file} {cmd.c_file}

"""
    return rules

def create_makefile(build_cmds):
    """
        description:
        ------------
            create makefile for build object files in user space using build_cmds list collected in collect_build_cmds()
        inputs:
        -------
            global build_cmds list
        outputs:
            makefile
    """
    objs_str=create_objs_str(build_cmds)
    compile_rules=create_compile_rules(build_cmds)
    kernel_implement_objs_str=""
    makefile_str=\
f"""\
objs={objs_str}
target=libcdev.so

all: $(target)
clean:
\t-rm $(target)
\t-rm $(objs)

$(target): $(objs)
\tgcc -shared -o $@ $(objs)

{compile_rules}

"""
    fd=open("Makefile.user", "w")
    fd.write(makefile_str)
    fd.close()

def main():
    current_dir=os.path.abspath(os.getcwd())
    build_cmds_kernel=collect_build_cmds(current_dir)
    build_cmds_user=modify_gcc_from_kernel_to_user(build_cmds_kernel)
    create_makefile(build_cmds_user)

if __name__ == "__main__":
    main()
